import logging

from ..settings import LOGGER_NAME
from time import perf_counter
from functools import wraps

log = logging.getLogger(LOGGER_NAME)
function_logs = {}

def timer(fnc):
    if fnc not in function_logs:
        function_logs[fnc] = {
            'name' : fnc.__name__,
            'time' : 0,
            'call_count': 0
        }
    
    @wraps(fnc)
    def inner(*args, **kwargs):
        function_logs[fnc]['call_count'] += 1
        t1 = perf_counter()
        result = fnc(*args, **kwargs)
        function_logs[fnc]['time'] += perf_counter() - t1
        return result
    return inner




import re

from .base import AioreqError


class UrlParsingError(AioreqError):

    def __init__(self, *args):
        text = "Can't handle this URL parsing error, please report your problem and help this project to be better"

        for i in range(1):
            url, = args
            if not (url.startswith('http') or url.startswith('https')):
                text = 'Url should starts with `http://` or `https://`.'
                break
            else:
                scheme_len = 7 if url.startswith('http://') else 8
                splited = url[scheme_len:]

                match = re.match(r'(.+?\.)?(?P<domain>.*?)\.(?P<top_domain>.*)[/$]', splited)

                if match:
                    domain, top_level_domain = bool(match.group('domain')), bool(match.group('top_domain'))
                    two_parts = (not domain) and (not top_level_domain)

                    missing_parts = ('Domain' if not domain else 'Top-level-domain') \
                        if not two_parts else ('Domain', 'Top-level-domain')

                    text = f"Missing url part{'s' if two_parts else ''}: {missing_parts}"

                    break
                else:
                    text = 'Unexpected domain syntax.'
                    break

        super(UrlParsingError, self).__init__(text)



import logging
from abc import abstractmethod, ABC

from enum import Enum
from ..settings import LOGGER_NAME
from ..parser.response_parser import ResponseParser

from typing import Tuple

log = logging.getLogger(LOGGER_NAME)


class ResponseParserStrategy(Enum):
    """
    An enumeration that implements the strategy design pattern used to
    choice of method for parsing the answer.
    """
    chunked = 'chunked'
    content_length = 'content_length'

    @staticmethod
    def parse_content_length(buffer: 'Buffer') -> bytes | None:
        """
        Parse incoming `BaseBuffer` object receiving data on which body length by `content-length` header.

        RFC[2616] 14.13 Content-Length:
            The Content-Lenght entity-header field indicates the size of the entity-body,
            in decimal number of OCTETs, sent to the recipent or, in the case of the HEAD method,
            the size of the entity-body that would have been sent had the request been a GET
        :param buffer: A buffer which data should be parsed
        :type buffer: BaseBuffer
        :return: Parsed and verified HTTP response or `NoneType` object
        """

        if len(buffer.text) >= buffer.content_length:
            buffer.switch_data(buffer.content_length)
            return buffer.message_verify()
        return None

    @staticmethod
    def parse_chunked(buffer: 'Buffer') -> None | bytes:
        """
        Parse incoming `BaseBuffer` object receiving data on which body length by chunked transfer encoding.

        RFC[2616] 3.6.1 Chunked Transfer Coding:
            The chunked encoding modifies the body of a message in order to transfer it as a series of
            chunkd, each with its own size indicator, followed by an OPTIONAL trailer containing entity-header
            fields. This allows dynamically produced content to be transferred along with the information
            necessary for the recipient to verify that it has received the full message
        :param buffer: A buffer which data should be parsed
        :type buffer: BaseBuffer
        :return: Parsed and verified HTTP response or `NoneType` object
        """

        while True:
            if buffer.bytes_should_receive_and_save:
                if buffer.bytes_should_receive_and_save <= len(buffer.text):
                    buffer.switch_data(buffer.bytes_should_receive_and_save)
                    buffer.bytes_should_receive_and_save = 0
                    buffer.bytes_should_receive_and_ignore = 2
                else:
                    break
            elif buffer.bytes_should_receive_and_ignore:
                if buffer.bytes_should_receive_and_ignore <= len(buffer.text):
                    buffer.ignore_data(buffer.bytes_should_receive_and_ignore)
                    buffer.bytes_should_receive_and_ignore = 0
                else:
                    break

            else:
                pattern = ResponseParser.regex_end_chunk
                end_match = pattern.search(buffer.text)
                if end_match:
                    return buffer.message_verify()

                match = ResponseParser.regex_find_chunk.search(buffer.text)
                if match is None or match.groups()[0] == b'0':
                    break
                size = int(match.group('content_size'), 16)
                buffer.bytes_should_receive_and_save = size
                buffer.ignore_data(match.end() - match.start())

    def parse(self, buffer: 'Buffer') -> bytes | None:
        """
        General interface to work with parsing strategies
        :param buffer: A buffer which data should be parsed
        :type buffer: BaseBuffer
        :returns: Parsed and verified HTTP response or `NoneType` object
        :rtype: bytes or None
        """

        match self.value:
            case 'content_length':
                return self.parse_content_length(buffer)
            case 'chunked':
                return self.parse_chunked(buffer)


class BaseBuffer:

    @abstractmethod
    def __init__(self):
        ...

    @abstractmethod
    def add_data(self, text: bytes) -> tuple[bytes, bool] | Tuple[bytes, int] | Tuple[None, None]:
        ...

    def set_up(self) -> None:
        self.__init__()


class Buffer(BaseBuffer):
    """
    Implementing message receiving using ResponseParserStrategy which support
    receiving by content_length or chunked
    """

    def __init__(self) -> None:

        self.text = bytearray()
        self.__headers_done: bool = False
        self.body_receiving_strategy: ResponseParserStrategy | None = None
        self.content_length: int | None = None
        self.bytes_should_receive_and_save: int = 0
        self.bytes_should_receive_and_ignore: int = 0
        self.message_data: bytearray = bytearray()
        self.without_body_len: int | None = None
        self.verified: bool = False

    def switch_data(self, length: int) -> None:
        """
        Deletes from buffer's unparsed data and add to parsed ones
        :param length: Message length to delete from unpares data
        :returns: None
        """

        for byte in self.text[:length]:
            self.message_data.append(byte)
        self.text = self.text[length:]

    def message_verify(self) -> bytes:
        """
        Should be called whenever response messages seem full.

        :returns: None
        """

        msg = self.message_data

        self.message_data = bytearray()
        self.verified = True
        return msg

    def ignore_data(self, length: int) -> None:
        """
        Removes the elements from the buffer by the given length
        :param length: Bytes count to remove from the buffer's unparsed data
        :returns: None
        """

        self.text = self.text[length:]

    def headers_done(self) -> bool:
        """
        Checks if the text contains HTTP message data including full headers
        or full headers are not received yet.
        """

        if not self.__headers_done:
            is_done = ResponseParser.headers_done(self.text)
            if is_done:
                without_body_len = ResponseParser.get_without_body_length(self.text)
                self.without_body_len = without_body_len
                self.switch_data(without_body_len)
            self.__headers_done = is_done
        return self.__headers_done

    def find_strategy(self) -> None:
        """
        Finds and sets a strategy for getting a message, which can be chunked
        or by content_length
        :returns: None
        """

        content_length = ResponseParser.search_content_length(self.message_data)
        if content_length is not None:
            self.content_length = content_length
            self.body_receiving_strategy = ResponseParserStrategy.content_length
        else:
            self.body_receiving_strategy = ResponseParserStrategy.chunked
        log.debug(f"Strategy found: {self.body_receiving_strategy}")

    def fill_bytes(self, _bytes: bytes):
        """
        Adds given bytes into the buffer's unparsed data
        :param _bytes: Bytes to add to the buffer
        :returns: None
        """
        for byte in _bytes:
            self.text.append(byte)

    def add_data(self, text: bytes) -> Tuple[bytes, int] | Tuple[None, None]:
        """
        Called whenever new data is required to be added
        :param text: text to add
        :type text: bytes
        :returns: `Tuple[None, None]` if the message not verified else full body and without body len
        :rtype: Tuple[bytes, int] | Tuple[None, None]
        """

        self.fill_bytes(text)

        if self.headers_done():

            if not self.body_receiving_strategy:
                self.find_strategy()

            result = self.body_receiving_strategy.parse(self)  # type: ignore
            if result:
                return result, self.without_body_len
        return None, None


class StreamBuffer(BaseBuffer, ABC):
    """
    The buffer for stream responses.

    This class is meant to work with big response data. If it receives all response data at once, the program can crash
    because of ram overflow. Therefore, handling each chunk separately is also supported.
    """

    def __init__(self):
        self.buffer = Buffer()
        self.headers_skipped = False

    def add_data(self,
                 text: bytes) -> tuple[bytes, bool]:
        """
        Adds new bytes into the buffer and gets parsed result if possible, otherwise returns `None`
        """

        done = self.buffer.add_data(text)

        if self.buffer.verified:
            if self.headers_skipped:
                return done[0], True
            return done[0][self.buffer.without_body_len:], True

        if self.buffer.without_body_len:
            if not self.headers_skipped:  # if headers not received, receive and clean the bytearray
                self.headers_skipped = True

                # if transfer encoding was chunked, then with headers it can also receive some chunks
                # This code receive additional content which comes with `headers`
                body_received = self.buffer.message_data[self.buffer.without_body_len:]
                self.buffer.message_data = bytearray(b'')  # clear message_data
                return body_received, False
        msg = self.buffer.message_data
        self.buffer.message_data = bytearray(b'')  # clear message_data
        return msg, False



import asyncio
import logging

from ..errors.requests import RequestTimeoutError
from typing import Awaitable
from ..settings import LOGGER_NAME
from ..settings import DEFAULT_TIMEOUT

log = logging.getLogger(LOGGER_NAME)


async def wrap_errors(coro: Awaitable, timeout: int):

    try:
        return await asyncio.wait_for(coro, timeout=timeout or DEFAULT_TIMEOUT)
    except asyncio.exceptions.TimeoutError:
        raise RequestTimeoutError(f"Request timeout error")
    except BaseException as e:
        log.critical(f"Unecpected error was raised")
        raise e



import logging

from pathlib import Path
from configparser import ConfigParser

ini_file_path_posix = Path(__file__).parent / 'settings.ini'
ini_file_path = str(ini_file_path_posix.absolute())

parser = ConfigParser()
parser.read(ini_file_path)

log_level_mapper = {
    'notset': logging.NOTSET,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}

log_format_mapper = {
    '$name': '%(name)s',
    '$levelno': '%(levelno)s',
    '$levelname': '%(levelname)s',
    '$pathname': '%(pathname)s',
    '$filename': '%(filename)s',
    '$module': '%(module)s',
    '$lineno': '%(lineno)d',
    '$funcName': '%(funcName)s',
    '$created': '%(created)f',
    '$asctime': '%(asctime)s',
    '$msecs': '%(msecs)d',
    '$relativeCreated': '%(relativeCreated)d',
    '$thread': "%(thread)d",
    '$threadName': '%(threadName)s',
    '$process': '%(process)d',
    '$message': '%(message)s'
}

LOGGER_NAME = parser.get('Logging', 'logger_name')

MAIN_LOGGER_LEVEL = parser.get('Logging', 'logger_level')
STREAM_HANDLER_LEVEL = parser.get('Logging', 'stream_handler_level')

REQUEST_REDIRECT_COUNT = parser.getint('Aioreq', 'request_redirect_count')
REQUEST_RETRY_COUNT = parser.getint('Aioreq', 'request_retry_count')

TEST_SERVER_DOMAIN = parser.get('Tests', 'test_server_domain')

DEFAULT_TIMEOUT = parser.getint('Aioreq', 'request_timeout')

if any(
        (
            (MAIN_LOGGER_LEVEL not in log_level_mapper),
            (STREAM_HANDLER_LEVEL not in log_level_mapper)
        )
):
    raise ValueError('Setting.ini contains invalid value '
                     f'for one of the logger levels ({MAIN_LOGGER_LEVEL} or {STREAM_HANDLER_LEVEL})')

MAIN_LOGGER_LEVEL = log_level_mapper.get(MAIN_LOGGER_LEVEL)  # type: ignore
STREAM_HANDLER_LEVEL = log_level_mapper.get(STREAM_HANDLER_LEVEL)  # type: ignore

FORMAT = parser.get('Logging', 'stream_handler_format')

for key, value in log_format_mapper.items():
    FORMAT = FORMAT.replace(key, value)

DEFAULT_CONNECTION_TIMEOUT = int(parser.getfloat('Connection', 'default_connection_timeout'))
DEFAULT_DNS_SERVER = parser['Connection']['default_dns_server']

main_logger = logging.getLogger(LOGGER_NAME)
main_logger.propagate = False
main_logger.setLevel(MAIN_LOGGER_LEVEL)

handler = logging.StreamHandler()
handler.setLevel(STREAM_HANDLER_LEVEL)

formatter = logging.Formatter(FORMAT)

handler.setFormatter(formatter)
main_logger.addHandler(handler)



__version__ = '1.0.0b1'

import asyncio

from .protocol import http
from .protocol.http import *
from .protocol import headers

try:
    ...
    # import uvloop

    # asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except:
    ...



# Informational

Continue = 100
SWITCHING_PROTOCOLS = 101

# Ok

OK = 200
CREATED = 201
ACCEPTED = 202
NON_AUTHORITATIVE_INFORMATION = 203
NO_CONTENT = 204
RESET_CONTENT = 205
PARTIAL_CONTENT = 206

# 3xx Redirects

MULTIPLE_CHOICES = 300
MOVED_PERMANENTLY = 301
FOUND = 302
SEE_OTHER = 303
NOT_MODIFIED = 304
USE_PROXY = 305
TEMPORARY_REDIRECT = 307

# 4xx Client errors

BAD_REQUEST = 400
UNAUTHORIZED = 401
PAYMENT_REQUIRED = 402
FORBIDDEN = 403
NOT_FOUND = 404
METHOD_NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
PROXY_AUTHENTICATION_REQUIRED = 407
REQUEST_TIMEOUT = 408
CONFLICT = 409



import zlib
import base64
import gzip as _gzip

from enum import Enum
from abc import ABC
from abc import abstractmethod



class Encoding(ABC):
    all_encodings = []

    @classmethod
    @abstractmethod
    def decompress(cls, text: bytes) -> bytes:
        ...

    def __init_subclass__(cls):
        Encoding.all_encodings.append(cls)

    @classmethod
    def stringify(cls) -> str:
        return cls.__name__


# Encodings
# -------------

class gzip(Encoding):

    @classmethod
    def decompress(cls, text: bytes) -> bytes:
        return _gzip.decompress(text)


class deflate(Encoding):

    @classmethod
    def decompress(cls, text: bytes) -> bytes:
        decoded_data = base64.b64decode(text)
        res = zlib.decompress(decoded_data, -15)
        return res
# class compress(Encoding):
#
#     @classmethod
#     def decompress(cls, text: bytes) -> bytes:
#         raise NotImplementedError


# Enum for encodings
# -------------

class Encodings(Enum):
    gzip = gzip
    deflate = deflate

    def decompress(self, text: bytes) -> bytes:
        return self.value.decompress(text)



from .base import AioreqError


class AsyncRequestsError(AioreqError): ...


class InvalidDomainName(AioreqError): ...


class RequestTimeoutError(AioreqError): ...


class ConnectionTimeoutError(AioreqError): ...



import re
import logging

from ..settings import LOGGER_NAME
from ..protocol.headers import TransferEncoding
from ..protocol.headers import ContentEncoding

from ..utils import debug

log = logging.getLogger(LOGGER_NAME)


class ResponseParser:
    """
    Used to parse raw response becoming from TCP connection
    """
    # Default regex to parse full response
    regex = re.compile(
        (
            r'(?P<scheme_and_version>HTTP/[210].[210]) (?P<status_code>\d{3}) (?P<status_message>.*?)\r\n'
            r'(?P<headers>[\d\D]*?)\r\n'
            r'\r\n'
            r'(?P<body>[\d\D]*)'
        ).encode(),
    )

    regex_without_body = re.compile(
        (
            r'(?P<scheme_and_version>HTTP/[210].[210]) (?P<status_code>\d{3}) (?P<status_message>.*?)\r\n'
            r'(?P<headers>[\d\D]*?)\r\n'
            r'\r\n'
        ).encode()
    )

    # Regex to find content-length header if exists
    regex_content = (r'\r\nContent-length\s*:\s*(?P<length>\d*)\r\n',
                     re.IGNORECASE)

    regex_content_length = re.compile(
        regex_content[0].encode(),
        regex_content[1]
    )

    regex_find_chunk = re.compile("^(?P<content_size>[0-9abcdefABCDEF]+)\r\n".encode())
    regex_end_chunk = re.compile(r'^0\r\n\r\n'.encode())


    @classmethod
    def parse_and_fill_headers(cls, binary_headers: bytes):
        headers = {}
        string_headers = binary_headers.decode()

        for line in string_headers.split('\r\n'):
            key, value = line.split(':', 1)
            headers[key.strip()] = value.strip()
        return headers

    @classmethod
    def decode_response_body(cls, response) -> None:
        for parser, header in (
                (TransferEncoding, 'transfer-encoding'),
                (ContentEncoding, 'content-encoding')
        ):
            header_content = response.headers.get(header, None)
            if header_content:
                encodings = parser.parse(header_content)

                for encoding in encodings:
                    response.content = encoding.decompress(response.content)

    @classmethod
    @debug.timer
    def parse(cls, response: bytes) -> 'Response':  # type: ignore
        """
        The main method for this class which parse response

        Parsing the raw response object and returning object type of
        Response which contains all becoming response data as his attributes
        :param response: raw response text
        :type response: bytes

        ..: todo: Remove
        """

        from ..protocol.http import Response
        match = cls.regex.search(response)
        scheme_and_version, status, status_message, unparsed_headers, body = match.groups()  # type: ignore
        headers = cls.parse_and_fill_headers(unparsed_headers)
        status = int(status)
        status_message = status_message.decode()

        response = Response(
            status=status,
            status_message=status_message,
            headers=headers,
            content=body
        )

        cls.decode_response_body(response)
        return response

    @classmethod
    @debug.timer
    def body_len_parse(cls, text: bytes, without_body_len: int):
        from ..protocol.http import Response

        withoutbody, body = text[:without_body_len], text[without_body_len:]

        match = cls.regex_without_body.search(withoutbody)
        scheme_and_version, status, status_message, unparsed_headers = match.groups()
        status = int(status)
        headers = cls.parse_and_fill_headers(unparsed_headers)

        response = Response(
            status=status,
            status_message=status_message.decode(),
            headers=headers,
            content=body
        )

        cls.decode_response_body(response)
        return response

    @classmethod
    @debug.timer
    def search_content_length(cls, text: bytes) -> int | None:
        """
        Search and returned content-length

        Search content-length header and return value if header exists
        using rexeg_content_length compiled regex
        :param text: text where content-length maybe exists
        :type text: bytes
        :returns: content_lenth | None
        :rtype: int or NoneType
        """

        match = cls.regex_content_length.search(text)
        if not match:
            return None
        content_length = match.group('length')
        return int(content_length)

    @classmethod
    @debug.timer
    def get_without_body_length(cls, text: bytes) -> int:
        """
        Get body less response

        Get index number from the text where body 
        part starting
        :param text: string where should be search
        :type text: bytes
        :returns: Index nuber where body part starts
        :rtype: int
        """

        match = cls.regex_without_body.match(text)
        assert match
        assert match.start() == 0, f"Got unexpected {match.start=}"
        return match.end() - match.start()

    @classmethod
    @debug.timer
    def headers_done(cls, text: bytes) -> bool:
        """
        Return true if text contains headers done text,
        which means HTTP message representing in string which
        contains an empty line
        """

        match = cls.regex_without_body.match(text)
        if match:
            assert match.start() == 0, f"Got unexpected {match.start=}"  # type: ignore
        return match is not None





import json as _json

from abc import abstractmethod
from typing import Iterable
from ..utils import debug


class BaseRequestParser:
    """
    For parsing Request object to raw data which can be sent
    via socket
    """

    @abstractmethod
    def parse(cls, request: 'Request') -> str: ...  # type: ignore

    @classmethod
    def sum_path_parameters(cls, parameters: Iterable[Iterable[str]]):
        return "&".join([f"{key}={value}" for key, value in parameters])


class RequestParser(BaseRequestParser):
    @classmethod
    @debug.timer
    def parse(cls, request: 'Request') -> str:  # type: ignore
        """
        Parsing object type of request to string representing HTTP message

        :returns: raw http request text
        :rtype: str
        """

        if request.path_parameters:
            request.path += '?' + \
                            cls.sum_path_parameters(request.path_parameters)

        if request.content:
            request.headers['Content-Length'] = len(request.content)

        message = (
                '\r\n'.join(
                    (
                        f'{request.method} {request.path} {request.scheme_and_version}',
                        f'host:  {request.host.split("://", 1)[1]}',
                        *request.headers.get_parsed(),
                    )
                ) + '\r\n\r\n')

        if type(request.content) in (bytes, bytearray):
            request.content = request.content.decode()
        message += request.content or ''

        return message


class JsonRequestParser(BaseRequestParser):

    @classmethod
    @debug.timer
    def parse(cls, request: 'Request') -> str:  # type: ignore
        """
        Parsing object type of request to string representing HTTP message

        :returns: raw http request text
        :rtype: str
        """

        if type(request.content) in (bytes, bytearray):
            request.content = request.content.decode()

        if request.path_parameters:
            request.path += '?' + \
                            cls.sum_path_parameters(request.path_parameters)

        if request.content:
            request.content = _json.dumps(request.content)[1:-1]
            request.headers['Content-Type'] = 'application/json'
            request.headers['Content-Length'] = len(request.content)

        message = ('\r\n'.join((
            f'{request.method} {request.path} {request.scheme_and_version}',
            f'host:  {request.host.split("://", 1)[1]}',
            *(f"{key}:  {value}" for key, value in request.headers.items()),
        )) + '\r\n\r\n')

        message += request.content or ''

        return message



import re
import logging

from dataclasses import dataclass
from ..errors.parser import UrlParsingError

from ..utils import debug

log = logging.getLogger('aioreq')


@dataclass
class Url:
    """
    Url class which stores url information

    This class used to store url information as attributes
    and gives interface to get specific part of that url

    :param protocol: Url protocol
    :type protocol: str or None
    :param subdomain: URL subdomain
    :type subdomain: str or None
    :param domain: URL domain
    :type domain: str or None
    :param top_level_domain: Top level domain like 'com' or 'ru' or 'am'
    :type top_level_domain: str or None
    :param path: Url path which comes after top level domain, like /users/id/132
    :type path: str or None
    :param variables: Variables which comes after path, like '?name=kar1&profession=programmer&group=&'
    :type variables: str or None
    :param fragment: Url fragment comes after varaibles like  #test where test is the fragment
    :type fragment: str or None
    """

    protocol: str | None
    subdomain: str | None
    domain: str
    top_level_domain: str
    path: str
    variables: str | None
    fragment: str | None

    def get_url(self):
        """
        Getting the full url

        :returns: full url
        :rtype: str
        """

        if self.subdomain:
            url = f"{self.protocol}://{self.subdomain}.{self.domain}.{self.top_level_domain}"
        else:
            url = f"{self.protocol}://{self.domain}.{self.top_level_domain}"
        url += self.path
        url += f'?{self.variables}' if self.variables else ''
        url += self.fragment or ''
        return url

    def get_url_without_path(self):
        """
        Getting pathless url

        :returns: pathless url
        :rtype: str
        """

        if self.subdomain:
            return f"{self.protocol}://{self.subdomain}.{self.domain}.{self.top_level_domain}"
        return f"{self.protocol}://{self.domain}.{self.top_level_domain}"

    def get_url_for_dns(self):
        """
        Getting only domains from the url

        :returns: joined domains with '.' char
        :rtype: str
        """

        if self.subdomain:
            return f"{self.subdomain}.{self.domain}.{self.top_level_domain}"
        return f"{self.domain}.{self.top_level_domain}"

    def __post_init__(self):
        """
        Function which dataclasses calls after __init__ done
        """

        self.path = self.path or '/'


class UrlParser:
    """
    Url parser which parse url and gives Url object
    """

    # regex which getting parts from the url
    regex = re.compile(
        r'(?P<protocol>https?)://'
        r'((?P<subdomain>[^\.]+)\.)?'
        r'(?P<domain>[^\.]+)\.'
        r'(?P<top_level_domain>[^/#]+)'
        r'(?P<port>\d+)?'
        r'(?:(?P<path>/[^#?]*)((?:\?'
        r'(?P<variables>[^#]*)?))?)?'
        r'(?:#(?P<fragment>.*))?'
    )

    @classmethod
    @debug.timer
    def parse(cls, url: str) -> Url:
        """
        The main function for this class, which parse url string

        Parsing raw url using regular expressions, creating Url object
        and returning

        :param url: url string
        :returns: Url object with attributes like url parts
        :rtype: Url
        """

        matched = cls.regex.search(url)

        if not matched:
            raise UrlParsingError(url)

        protocol = matched.group('protocol')
        subdomain = matched.group('subdomain')
        domain = matched.group('domain')
        top_level_domain = matched.group('top_level_domain')
        path = matched.group('path')
        variables = matched.group('variables')
        fragment = matched.group('fragment')
        return Url(
            protocol,
            subdomain,
            domain,
            top_level_domain,
            path,
            variables,
            fragment
        )



import asyncio
import logging
from abc import ABCMeta, ABC
from abc import abstractmethod
from collections import defaultdict
from typing import Any
from typing import AsyncIterable
from typing import Iterable
from typing import Union

from .encodings import Encoding
from .headers import AcceptEncoding
from .headers import BaseHeader
from ..errors.requests import ConnectionTimeoutError
from ..parser.request_parser import JsonRequestParser
from ..parser.request_parser import RequestParser
from ..parser.response_parser import ResponseParser
from ..parser.url_parser import Url
from ..parser.url_parser import UrlParser
from ..settings import LOGGER_NAME
from ..settings import REQUEST_REDIRECT_COUNT
from ..settings import REQUEST_RETRY_COUNT
from ..settings import TEST_SERVER_DOMAIN
from ..transports.connection import Transport
from ..transports.connection import resolve_domain
from ..utils import debug
from ..utils.generic import wrap_errors

log = logging.getLogger(LOGGER_NAME)


class HttpProtocol(ABC):
    """
    An abstract class for all Http units representing HTTP/1.1 protocol
    with the general attributes
    """

    safe_methods = (
        "GET",
        "HEAD"
    )

    scheme = 'HTTP'
    scheme_and_version = 'HTTP/1.1'
    version = '1.1'


class MetaHeaders(type):

    def __call__(cls, initial_headers=None):
        """
        If headers is a Header object, then don't create a new one.
        """
        if isinstance(initial_headers, Headers):
            return initial_headers
        return super(MetaHeaders, cls).__call__(initial_headers)


class Headers(metaclass=MetaHeaders):
    """
    Dict like object used to represent the HTTP headers.
    """

    def __init__(self,
                 initial_headers: Union[dict[str, str], None] = None):
        self._headers = {}
        self.cache = ''

        if initial_headers:
            for key, value in initial_headers.items():
                self[key] = value

    def __setitem__(self, key: str, value: str):
        self.cache = None
        self._headers[key.lower()] = value

    def __getitem__(self, item):
        return self._headers[item.lower()]

    def add_header(self, header: BaseHeader):
        self[header.key] = header.value

    def get_parsed(self):
        if self.cache is not None:
            return self.cache

        headers = (f"{key}:  {value}" for key, value in self._headers.items())
        self.cache = headers
        return headers

    def items(self):
        return self._headers.items()

    def get(self, name, default=None):
        return self._headers.get(name.lower(), default)

    def __contains__(self, item):
        return item.lower() in self._headers

    def __or__(self, other):
        if not isinstance(other, Headers):
            raise ValueError(f"Can't combine {self.__class__.__name__} object with {type(other).__name__}")

        return Headers(
            initial_headers=self._headers | other._headers
        )

    def __len__(self):
        return len(self._headers)

    def __repr__(self):
        return f"Headers:\n" + '\n'.join(
            (f" {key}: {value}" for key, value in self._headers.items())
        )


class BaseRequest(HttpProtocol, metaclass=ABCMeta):
    """
    An abstract class for all HTTP request classes

    This is a low level Request abstraction class which used by Client by default,
    but can be used directly.
    By default, used by 'aioreq.protocol.http.Client.send_request'
    :param method: HTTP method (GET, POST, PUT, PATCH),
    see also RFC[2616] 5.1.1 Method
    :type method: str
    :param headers: HTTP headers paired by (key, value) where 'key'= header name
    and 'value'= header value
    :type headers: dict[str, str]

    :Example:

    >>> from aioreq.protocol.http import Request
    >>> req = Request(
    ...              method='GET',
    ...              host='https://google.com',
    ...              path='/',
    ...              headers={})
    >>> print(req)
    "Request(GET, https://google/com)"
    """

    parser = None

    __slots__ = (
        'method',
        '_host',
        'headers',
        'path',
        'raw_request',
        'content',
        'path_parameters',
    )

    def __init__(
            self,
            url: str,
            *,
            headers: Union[Headers, dict[str, str], None] = None,
            method: str = 'GET',
            raw_request: Union[None, bytes] = None,
            content: Union[str, bytearray, bytes] = '',
            params: Union[Iterable[Iterable[str]], None] = None,
    ) -> None:
        """
        Request initialization method
        """

        if params is None:
            params = ()

        splited_url = UrlParser.parse(url)

        self._host = splited_url.get_url_without_path()
        self.path = splited_url.path

        self.headers = Headers(headers)

        self.method = method
        self.content = content
        self.path_parameters = params
        self._raw_request = raw_request

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, new_value):
        self._host = new_value.strip('/')

    def get_raw_request(self) -> bytes:
        """
        The Getter method for raw_request private attribute if
        request has been pares once, otherwise parse and return

        :returns: raw request for this Request abstraction
        :rtype: bytes
        """

        if self._raw_request:
            return self._raw_request

        message = self.parser.parse(self)
        enc_message = message.encode('utf-8')
        self._raw_request = enc_message
        return enc_message

    def __repr__(self) -> str:
        return f"<Request {self.method} {self.host}>"

    def __getattribute__(self, item):
        self._raw_request = None  # Clear cache
        return super().__getattribute__(item)


class BaseResponse(HttpProtocol, metaclass=ABCMeta):
    """
    An abstract Response class
    """

    __slots__ = (
        'status',
        'status_message',
        'headers',
        'content',
        'request'
    )


class Request(BaseRequest):
    parser = RequestParser


class JsonRequest(BaseRequest):
    parser = JsonRequestParser


class Response(BaseResponse):
    """
    An HTTP response asbtraction

    This is a Response abstraction class used by 'aioreq.parser.response_parser.ResponseParser.parse'
    by default to make response binary data more friendly and not reccommended to use directly
    :param status: response code returned with response,
    see also RFC[2616] 6.1.1 Status Code and Reason Pharse
    :type status: int
    :param status_message: description for :status: in the status line
    :type status_message: str
    :param headers: HTTP headers sent by the server, see also RFC[2616] 6.2
    Response Headers Fields
    :type headers: dict[str, str]
    :param content: response body
    :type content: bytes
    :param request: Request for this response 
    :type request: Request

    :Example:

    >>> import aioreq
    >>> from aioreq.protocol.http import Response
    >>> 
    >>> a = Response(
    ...          status=200,
    ...          status_message='OK',
    ...          headers={},
    ...          content=b'Test message',
    ...          request=None)
    >>> print(a)
    "Response(200, OK)" 
    """

    def __init__(
            self,
            status: int,
            status_message: str,
            headers: dict[str, str],
            content: bytes,
            request: Union[Request, None] = None):
        """
        Response initialization method
        """

        self.status = status
        self.status_message = status_message
        self.headers = Headers(headers)
        self.content = content
        self.request = request

    def __eq__(self, value: 'Response') -> bool:
        """
        Check if two Response objects have the same attributes or not
        :param value: right side value of equal
        :type value: Response 
        :returns: True if values are equal
        :rtype: bool
        """

        if type(self) != type(value):
            return False
        return self.__dict__ == value.__dict__

    def __repr__(self) -> str:
        return f"<Response {self.status} {self.status_message}>"


class BaseClient(metaclass=ABCMeta):
    """
       A session like class Client used to make requests

       This is a Client abstraction used to communicate with the
       HTTP protocol, send requests, receive responses and so on
       :param headers: HTTP protocol headers
       :type headers: dict[str, str], None
       :param persistent_connections: Persistent connections support for our client
       see also RFC[2616] 8.1 Persistent Connections
       :type persistent_connections: bool
       Header objects defined in 'aioreq.protocol.headers',
       this is an easy way to use HTTP Headers through OOP
       """

    @abstractmethod
    async def _send_request(self,
                            url: str,
                            method: str,
                            content: Union[str, bytearray, bytes] = '',
                            path_parameters: Union[Iterable[Iterable[str]], None] = None,
                            headers: Union[None, dict[str, str], Headers] = None,
                            ) -> Response:
        ...

    @abstractmethod
    async def send_request(self,
                           request: Request, ) -> Response:
        ...

    @staticmethod
    def get_avaliable_encodings():
        return AcceptEncoding(

            *((encoding, 1) for encoding in Encoding.all_encodings)

        )

    def __init__(self,
                 headers: Union[dict[str, str], Headers, None] = None,
                 persistent_connections: bool = False,
                 enable_encodings: bool = True):

        self.connection_mapper = defaultdict(list)

        headers = Headers(initial_headers=headers)

        if enable_encodings and 'accept-encoding' not in headers:
            accept_encoding_object = self.get_avaliable_encodings()
            headers.add_header(accept_encoding_object)

        self.headers = headers
        self.transports = []
        self.persistent_connections = persistent_connections

    async def _get_connection(self, splitted_url: Url):
        """
        Getting connection from already opened connections, to perform Keep-Alive logic,
        if these connections exists or create the new one and save into connection pool
        :param splitted_url: Url object which contains all url parts
        (protocol, version, subdomain, domain, ...)
        :type splitted_url: Url
        :returns: Transport
        """
        transport = None
        if self.persistent_connections:
            log.debug(f"{self.connection_mapper} searching into mapped connections")

            for transport in self.connection_mapper[splitted_url.get_url_for_dns()]:
                if not transport.used:
                    if transport.is_closing():
                        del self.connection_mapper[splitted_url.get_url_for_dns()]
                    else:
                        break
            else:
                transport = None

        if not transport:
            if splitted_url.domain == TEST_SERVER_DOMAIN:  # server for tests
                ip, port = 'localhost', 7575
            else:
                ip = await resolve_domain(splitted_url.get_url_for_dns())
                port = 443 if splitted_url.protocol == 'https' else 80

            transport = Transport()
            connection_coroutine = transport.make_connection(
                ip,
                port,
                ssl=splitted_url.protocol == 'https'
            )
            try:
                await connection_coroutine
                self.transports.append(transport)

            except asyncio.exceptions.TimeoutError as err:
                raise ConnectionTimeoutError('Socket connection timeout') from err

            if self.persistent_connections:
                self.connection_mapper[splitted_url.get_url_for_dns(
                )].append(transport)
        else:
            log.info("Using already opened connection")
        return transport

    async def __aenter__(self):
        """
        Implements 'with', closes transports after
        session ends

        :returns: Client object
        :type: Client
        """
        return self

    async def __aexit__(self, *args, **kwargs):
        """
        Closes using recourses

        :returns: None
        """
        for fnc, log_data in debug.function_logs.items():
            time = log_data['time']
            call_count = log_data['call_count']
            log.debug(f"Function {fnc.__module__}::{fnc.__name__} log | exec time: {time} | call count: {call_count}")

        tasks = []

        for transport in self.transports:
            transport.writer.close()
            log.debug('Trying to close')
            tasks.append(transport.writer.wait_closed())
        await asyncio.gather(*tasks)


class Client(BaseClient):
    """
    :Example:

    >>> import aioreq
    >>> import asyncio
    ...
    >>> async def main():
    >>>     async with aioreq.http.Client() as cl:
    >>>         return await cl.get('https://www.youtube.com')
    >>> resp = asyncio.run(main())

    .. todo: Isolate clients utils.debug.timer logging system
    """

    async def get(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[dict[str, str], None] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0,
            redirect: int = REQUEST_REDIRECT_COUNT,
            retry: int = REQUEST_RETRY_COUNT) -> Response:
        return await self.request_retry_wrapper(
            url=url,
            method="GET",
            content=content,
            headers=headers,
            path_parameters=path_parameters,
            timeout=timeout,
            redirect=redirect + 1,
            retry=retry + 1
        )

    async def post(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[dict[str, str], None] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0,
            redirect: int = REQUEST_REDIRECT_COUNT,
            retry: int = REQUEST_RETRY_COUNT) -> Response:
        return await self.request_retry_wrapper(
            url=url,
            method="POST",
            content=content,
            headers=headers,
            path_parameters=path_parameters,
            timeout=timeout,
            redirect=redirect + 1,
            retry=retry + 1
        )

    async def put(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[dict[str, str], None] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0,
            redirect: int = REQUEST_REDIRECT_COUNT,
            retry: int = REQUEST_RETRY_COUNT) -> Response:
        return await self.request_retry_wrapper(
            url=url,
            method="PUT",
            content=content,
            headers=headers,
            path_parameters=path_parameters,
            timeout=timeout,
            redirect=redirect + 1,
            retry=retry + 1
        )

    async def delete(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[dict[str, str], None] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0,
            redirect: int = REQUEST_REDIRECT_COUNT,
            retry: int = REQUEST_RETRY_COUNT) -> Response:
        return await self.request_retry_wrapper(
            url=url,
            method="DELETE",
            content=content,
            headers=headers,
            path_parameters=path_parameters,
            timeout=timeout,
            redirect=redirect + 1,
            retry=retry + 1
        )

    async def options(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[dict[str, str], None] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0,
            redirect: int = REQUEST_REDIRECT_COUNT,
            retry: int = REQUEST_RETRY_COUNT) -> Response:
        return await self.request_retry_wrapper(
            url=url,
            method="OPTIONS",
            content=content,
            headers=headers,
            path_parameters=path_parameters,
            timeout=timeout,
            redirect=redirect + 1,
            retry=retry + 1
        )

    async def head(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[dict[str, str], None] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0,
            redirect: int = REQUEST_REDIRECT_COUNT,
            retry: int = REQUEST_RETRY_COUNT) -> Response:
        return await self.request_retry_wrapper(
            url=url,
            method="HEAD",
            content=content,
            headers=headers,
            path_parameters=path_parameters,
            timeout=timeout,
            redirect=redirect + 1,
            retry=retry + 1
        )

    async def patch(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[dict[str, str], None] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0,
            redirect: int = REQUEST_REDIRECT_COUNT,
            retry: int = REQUEST_RETRY_COUNT) -> Response:
        return await self.request_retry_wrapper(
            url=url,
            method="PATCH",
            content=content,
            headers=headers,
            path_parameters=path_parameters,
            timeout=timeout,
            redirect=redirect + 1,
            retry=retry + 1
        )

    async def send_request_directly(self,
                                    request: Request,
                                    timeout: int = 0):

        splited_url = UrlParser.parse(request.host.strip('/') + request.path)
        transport = await self._get_connection(splited_url)
        coro = transport.send_http_request(request.get_raw_request())
        raw_response, without_body_len = await wrap_errors(coro=coro, timeout=timeout)
        resp = ResponseParser.body_len_parse(raw_response, without_body_len)
        resp.request = request
        return resp

    async def _send_request(self,
                            url: str,
                            method: str,
                            content: Union[str, bytearray, bytes] = '',
                            path_parameters: Union[Iterable[Iterable[str]], None] = None,
                            headers: Union[None, dict[str, str]] = None,
                            timeout: int = 0) -> Response:

        """
        Simulates a http request
        :param url: Url where should be request send
        :type url: str
        :param headers: Http headers which should be used in this GET request
        :type headers: Dict[str, str]
        :param content: Http body part
        :type content: str or bytearray or bytes
        :param method: Http message method
        :type method: str
        :param path_parameters:
        :type path_parameters: Iterable[BaseHeader] or None
        :param timeout: The request timeout
        :type timeout: int
        :returns: Response object which represents returned by server response
        :rtype: Response
        """

        headers = Headers(initial_headers=headers)

        splitted_url = UrlParser.parse(url)
        transport = await self._get_connection(splitted_url)
        request = Request(
            url=url,
            method=method,
            headers=self.headers | headers,
            params=path_parameters,
            content=content,
        )
        coro = transport.send_http_request(request.get_raw_request())
        raw_response, without_body_len = await wrap_errors(coro=coro, timeout=timeout)
        resp = ResponseParser.body_len_parse(raw_response, without_body_len)
        resp.request = request
        return resp

    async def request_redirect_wrapper(self,
                                       *args: tuple[Any],
                                       redirect: int,
                                       **kwargs,
                                       ) -> Response:
        """
        A wrapper method for send_request, also implements redirection if
        3xx status code received
        :param redirect: Maximum request sending counts
        :type redirect: int
        :return: Response object
        :rtype: Response
        """
        request = kwargs.get('request')

        redirect = max(1, redirect)  # minimum one request required

        while redirect != 0:
            redirect -= 1
            if request:
                result = await self.send_request_directly(request)
            else:
                result = await self._send_request(*args, **kwargs)

            if (result.status // 100) == 3:

                if not request:
                    kwargs['url'] = result.headers['Location']
                else:
                    request.host = result.headers['Location']

                if redirect < 1:
                    logging.debug(result, 'last one')
                    return result
            else:
                logging.debug(result)
                return result
            log.info(f'Redirecting request with status code {result.status}')

    async def request_retry_wrapper(self,
                                    *args: tuple[Any],
                                    retry: int,
                                    **kwargs
                                    ) -> Response:
        """
        A wrapper method for request_redirect_wrapper, also implements retrying
        for the requests if they were failed
        :param retry: Maximum request sending count
        :type retry: int
        :return: Response object
        :rtype: Response
        """

        retry = max(1, retry)  # minimum one request required

        while retry != 0:
            retry -= 1
            try:
                result = await self.request_redirect_wrapper(*args, **kwargs)
                return result
            except BaseException as e:
                if retry < 1:
                    raise e
                log.info(f'Retrying request cause of {e}')

    async def send_request(self,
                           request: BaseRequest,
                           timeout: int = 0,
                           redirect: int = REQUEST_REDIRECT_COUNT,
                           retry: int = REQUEST_RETRY_COUNT) -> Response:
        """
        Send request by giving Request object directly
        :param request: Request instance
        :type request: Request
        :param timeout: Request timeout
        :type timeout: int
        :param redirect: Request maximum redirect count
        :type redirect: int
        :param retry: Request retry count
        :type retry: int

        :returns: Response
        :rtype Response:

        """
        return await self.request_retry_wrapper(
            retry=retry,
            redirect=redirect,
            request=request,
            timeout=timeout,
        )


class StreamClient(BaseClient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, enable_encodings=False)

    async def _send_request(self,
                            url: str,
                            method: str,
                            content: Union[str, bytearray, bytes] = '',
                            path_parameters: Union[Iterable[Iterable[str]], None] = None,
                            headers: Union[None, dict[str, str]] = None,
                            timeout: int = 0) -> AsyncIterable:
        headers = Headers(initial_headers=headers)

        request = Request(
            url=url,
            method=method,
            headers=self.headers | headers,
            params=path_parameters,
            content=content,
        )
        async for chunk in self.send_request(request):
            yield chunk

    async def post(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[None, dict[str, str]] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0):
        async for chunk in self._send_request(
                url=url,
                method="POST",
                content=content,
                headers=headers,
                path_parameters=path_parameters,
                timeout=timeout,
        ):
            yield chunk

    async def get(
            self,
            url: str,
            content: Union[str, bytearray, bytes] = '',
            headers: Union[None, dict[str, str]] = None,
            path_parameters: Union[Iterable[Iterable[str]], None] = None,
            timeout: int = 0):
        async for chunk in self._send_request(
                url=url,
                method="GET",
                content=content,
                headers=headers,
                path_parameters=path_parameters,
                timeout=timeout,
        ):
            yield chunk

    async def send_request(self,
                           request: BaseRequest) -> AsyncIterable:
        splited_url = UrlParser.parse(request.host + request.path)
        transport = await self._get_connection(splited_url)
        coro = transport.send_http_stream_request(request.get_raw_request())
        async for chunk in coro:
            yield chunk



"""
Contains Header classes to simplify header sending
"""

from abc import ABC
from abc import abstractmethod
from collections.abc import Collection
from enum import Enum

from typing import Tuple

from .encodings import Encoding
from .encodings import Encodings


def qvalue_validate(qvalue: int) -> bool:
    if 0 <= qvalue <= 1:
        return True
    return False


class MimeType(Enum):
    json = 'application/json'
    html = 'application/html'


class ServerHeader(ABC):

    @classmethod
    @abstractmethod
    def parse(self, value: str) -> str: ...


class BaseHeader(ABC):
    key = 'NotImplemented'

    @property
    @abstractmethod
    def value(self) -> str: ...


class AcceptEncoding(BaseHeader):
    """
    RFC[2616] 14.3
        Accept-Encoding request-header field is similar to Accept, but
        restricts the content-codings (section 3.5) that are acceptable in
        the response.
    """

    key = 'Accept-Encoding'

    def __init__(
            self,
            *codings: Tuple[
                Tuple[Encoding, int]
                | Tuple[Encoding]
                ]):
        self._codings = {}
        for coding in codings:
            assert 0 < len(coding) < 3
            if len(coding) == 2:
                coding_type, qvalue = coding  # type: ignore
            else:
                coding_type = coding[0]
                qvalue = 1

            if not qvalue_validate(qvalue):
                raise ValueError("Invalid qvalue given -> {qvalue}. Expected int between 0, 1")

            if isinstance(coding_type, Encodings):
                coding_type = coding_type.value
            self._codings[coding_type.stringify()] = qvalue

    @property
    def value(self):
        text = ' '
        for coding, qvalue in self._codings.items():
            text += f"{str(coding)}; q={qvalue}, "
        if self._codings:
            text = text[:-2]
        return text


class Accept(BaseHeader):
    """
    RFC[2616] 14.1
        The Accept request-header field can be used to specify certain media
        types which are acceptable for the response. Accept headers can be
        used to indicate that the request is specifically limited to a small
        set of desired types, as in the case of a request for an in-line image.
    """

    key = 'Accept'

    def __init__(
            self,
            *types: Tuple[
                tuple[
                    MimeType,
                    int | None
                ] |
                tuple[
                    MimeType
                ]
                ]
    ):

        self.media_ranges = {}
        for media_range in types:
            assert 0 < len(media_range) < 3
            if len(media_range) == 2:
                type, qvalue = media_range
            else:
                type, = media_range
                qvalue = 1
            self.media_ranges[type] = qvalue

    @property
    def value(self) -> str:
        text = ' '
        for type, qvalue in self.media_ranges.items():
            text += f'{type.value}; q={qvalue}, '
        text = text[:-2]
        return text


class ServerEncoding(ServerHeader):

    def __init__(self):
        self.encodings = []
        self.iternum = 0

    @classmethod
    def parse(cls, text: str):
        self = cls()
        encodings = text.split(',')
        for encoding in encodings:
            if encoding != 'chunked':
                self.encodings.append(Encodings[encoding.strip()])
        self.encodings.reverse()
        return self

    def __iter__(self):
        self.iternum = 0
        return self

    def __next__(self):
        if self.iternum >= len(self.encodings):
            raise StopIteration
        encoding = self.encodings[self.iternum]
        self.iternum += 1
        return encoding


class TransferEncoding(ServerEncoding): ...


class ContentEncoding(ServerEncoding): ...



import asyncio
import logging
import os
import ssl
from typing import Tuple

import certifi
from dns import asyncresolver

from .buffer import Buffer, StreamBuffer
from ..errors.requests import AsyncRequestsError
from ..settings import LOGGER_NAME

res = asyncresolver.Resolver()
res.nameservers = ['1.1.1.1', '8.8.8.8']

log = logging.getLogger(LOGGER_NAME)

context = ssl.SSLContext(ssl.PROTOCOL_TLS)
context.load_verify_locations(certifi.where())
context.keylog_filename = os.getenv('SSLKEYLOGFILE')


async def get_address(host):
    answers = await res.resolve(host)
    return answers.rrset[0].address


dns_cache = dict()


async def resolve_domain(
        hostname: str,
) -> str:
    """
    Makes an asynchronous DNS request to resolve the IP address
    :param hostname: Domain name for example YouTube.com
    :type hostname: str
    :returns: ip and port for that domain
    :rtype: [str, int]
    """
    if hostname in dns_cache:
        if not isinstance(dns_cache[hostname], str):
            log.debug('Got cached dns query')
            return await dns_cache[hostname]
        return dns_cache[hostname]

    log.debug(f"trying resolve {hostname=}")
    coro = asyncio.create_task(get_address(hostname))
    dns_cache[hostname] = coro
    host = await coro
    dns_cache[hostname] = host
    return host


class Transport:
    """
    An asynchronous sockets communication implementation using
    asyncio streams.
    """

    def __init__(self):
        self.reader: asyncio.StreamReader | None = None
        self.writer: asyncio.StreamWriter | None = None
        self.used: bool = False
        self.message_manager: Buffer = Buffer()
        self.stream_message_manager: StreamBuffer = StreamBuffer()

    async def _send_data(self, raw_data: bytes) -> None:
        """
        An asynchronous alternative for socket.send_all method
        :param raw_data: Data which should be sent
        :type raw_data: bytes

        :returns: None
        """
        self.writer.write(raw_data)
        await self.writer.drain()

    async def _receive_data(self) -> Tuple[None, None] | Tuple[bytes, int]:
        """
        An asynchronous alternative for socket.recv() method.
        """
        data = await self.reader.read(1000)
        resp, without_body_len = self.message_manager.add_data(data)
        return resp, without_body_len

    async def _receive_data_stream(self):
        data = await self.reader.read(20)
        headerless_data = self.stream_message_manager.add_data(data)
        return headerless_data

    async def make_connection(
            self,
            ip: str,
            port: int,
            ssl: bool) -> None:
        """
        An asynchronous alternative for socket connect
        :param ip: Ip where connection should be made
        :type ip: str
        :param port: Connection port
        :type port: int
        :param ssl: True if TLS/SSL should be used
        :type ssl: bool
        :returns: None
        """

        reader, writer = await \
            asyncio.wait_for(asyncio.open_connection(
                host=ip,
                port=port,
                ssl=context if ssl else None
            ),
                timeout=3
            )
        self.reader = reader
        self.writer = writer

    def _check_used(self):
        if self.used:
            raise AsyncRequestsError('Using transport which is already in use')
        self.used = True

    async def send_http_request(self, raw_data: bytes) -> Tuple[bytes, int]:
        """
        The lowest level http request method, can be used directly
        :param raw_data: HTTP message bytes
        :type raw_data: bytes

        :returns: Response bytes and without data len
        :rtype: Tuple[bytes, int]
        """
        self._check_used()
        self.message_manager.set_up()
        await self._send_data(raw_data)
        while True:

            resp, without_body_len = await self._receive_data()
            if resp is not None:
                self.used = False
                return resp, without_body_len

    async def send_http_stream_request(
            self,
            raw_data: bytes):

        self._check_used()
        self.stream_message_manager.set_up()
        await self._send_data(raw_data)

        while True:
            body, done = await self._receive_data_stream()
            if body:
                yield body
            if done:
                break
            await asyncio.sleep(0)

    def is_closing(self) -> bool:
        """
        Wraps transport is_closing
        """
        return self.writer.is_closing()

    def __repr__(self):
        return f"<Transport {'Closed' if self.is_closing() else 'Open'}>"
